/*  $Id: FrenchVerbDictionary.cpp,v 1.41 2010/08/12 03:50:05 sarrazip Exp $
    FrenchVerbDictionary.cpp - Dictionary of verbs and conjugation templates

    verbiste - French conjugation system
    Copyright (C) 2003-2010 Pierre Sarrazin <http://sarrazip.com/>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.
*/

#include "FrenchVerbDictionary.h"

#include <iostream>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

using namespace std;
using namespace verbiste;


static bool trace = getenv("TRACE") != NULL;


class AutoDoc
{
public:
    AutoDoc(xmlDocPtr d) : doc(d) {}
    ~AutoDoc() { if (doc != NULL) xmlFreeDoc(doc); }
    xmlDocPtr get() const { return doc; }
    bool operator ! () const { return doc == NULL; }
private:
    xmlDocPtr doc;

    // Forbidden operations:
    AutoDoc(const AutoDoc &);
    AutoDoc &operator = (const AutoDoc &);
};


class AutoString
{
public:
    AutoString(xmlChar *s) : str(s) {}
    ~AutoString() { if (str != NULL) xmlFree(str); }
    xmlChar *get() const { return str; }
    bool operator ! () const { return str == NULL; }
    size_t length() const { return str == NULL ? 0 : strlen((char *) str); }
private:
    xmlChar *str;

    // Forbidden operations:
    AutoString(const AutoString &);
    AutoString &operator = (const AutoString &);
};


inline
const xmlChar *
XMLCHAR(const char *s)
{
    return (const xmlChar *) s;
}


inline
int
equal(const xmlChar *a, const char *b)
{
    return xmlStrcmp(a, XMLCHAR(b)) == 0;
}


inline
int
different(const xmlChar *a, const char *b)
{
    return !equal(a, b);
}


inline
xmlChar *
getProp(xmlNodePtr node, const char *propName)
{
    return xmlGetProp(node, XMLCHAR(propName));
}


inline
xmlChar *
getString(xmlDocPtr doc, xmlNodePtr node)
{
    return xmlNodeListGetString(doc, node, 1);
}


inline
string
operator + (const AutoString &a, const string &b)
{
    return (char *) a.get() + b;
}


inline
string
operator + (const string &a, const AutoString &b)
{
    return a + (char *) b.get();
}


inline
Mode
convertModeName(const xmlChar *modeName)
{
    return FrenchVerbDictionary::convertModeName((char *) modeName);
}


inline
Tense
convertTenseName(const xmlChar *tenseName)
{
    return FrenchVerbDictionary::convertTenseName((char *) tenseName);
}


// Latin-1 to ASCII conversion table (codes 0xC0 to 0xFF).
// Some characters have bogus translations, but they are not used in French.
//
static const char *accentRemovalTable =
            "AAAAAA_CEEEEIIII"
            "DNOOOOOxOUUUUYbB"
            "aaaaaa-ceeeeiiii"
            "dnooooo/ouuuuyby";


static char
removeAccent(char c)
{
    unsigned char u = static_cast<unsigned char>(c);
    if (u >= 0xC0)
        u = accentRemovalTable[u - 0xC0];
    return static_cast<char>(u);
}


void
FrenchVerbDictionary::formUnaccentedVariants(const string &latin1String,
                                                size_t index,
                                                vector<string> &variants)
{
    for ( ; index < latin1String.length(); ++index)
    {
        char ch = latin1String[index];
        char unacc = removeAccent(ch);
        if (ch != unacc)
        {
            string copy = latin1String;
            copy[index] = unacc;
            variants.push_back(copy);
            formUnaccentedVariants(copy, index + 1, variants);
        }
    }
}


string
FrenchVerbDictionary::removeAccents(const string &latin1String)
{
    string result;
    result.reserve(latin1String.length());
    for (size_t i = 0; i < latin1String.length(); ++i)
        result += removeAccent(latin1String[i]);
    return result;
}


void
FrenchVerbDictionary::getXMLFilenames(string &conjFN, string &verbsFN,
                                      const string &languageCode)
{
    const char *libdatadir = getenv("LIBDATADIR");
    if (libdatadir == NULL)
        libdatadir = LIBDATADIR;
    conjFN  = libdatadir + string("/") + "conjugation-" + languageCode + ".xml";
    verbsFN = libdatadir + string("/") + "verbs-" + languageCode + ".xml";
}


FrenchVerbDictionary::FrenchVerbDictionary(
                                const string &conjugationFilename,
                                const string &verbsFilename,
                                bool includeWithoutAccents,
                                const std::string &_langCode /*= "fr"*/)
                                        throw (logic_error)
  : conjugSys(),
    knownVerbs(),
    aspirateHVerbs(),
    inflectionTable(),
    toUTF8(),
    toLatin1(),
    verbTrie(*this),
    langCode(_langCode)
{
    if (langCode != "fr" && langCode != "it")
        throw logic_error("Invalid language code");
    init(conjugationFilename, verbsFilename, includeWithoutAccents);
}


FrenchVerbDictionary::FrenchVerbDictionary(bool includeWithoutAccents)
                                                throw (std::logic_error)
  : conjugSys(),
    knownVerbs(),
    aspirateHVerbs(),
    inflectionTable(),
    toUTF8(),
    toLatin1(),
    verbTrie(*this),
    langCode("fr")
{
    string conjFN, verbsFN;
    getXMLFilenames(conjFN, verbsFN, langCode);

    init(conjFN, verbsFN, includeWithoutAccents);
}


void
FrenchVerbDictionary::init(const string &conjugationFilename,
                            const string &verbsFilename,
                            bool includeWithoutAccents)
                                        throw (logic_error)
{
    toUTF8 = iconv_open("UTF-8", "ISO-8859-1");
    if (toUTF8 == (iconv_t) -1)
        throw logic_error("conversion from ISO-8859-1 to UTF-8 not supported");
    toLatin1 = iconv_open("ISO-8859-1", "UTF-8");
    if (toLatin1 == (iconv_t) -1)
        throw logic_error("conversion from UTF-8 to ISO-8859-1 not supported");

    {
        for (int i = 0; i < 0xC0; i++)
            latin1TolowerTable[i] = char(tolower(char(i)));
        for (int i = 0xC0; i < 0xE0; i++)
            latin1TolowerTable[i] = char(i + 0x20);
        for (int i = 0xE0; i < 0x100; i++)
            latin1TolowerTable[i] = char(i);
    }

    loadConjugationDatabase(conjugationFilename.c_str(), includeWithoutAccents);
    loadVerbDatabase(verbsFilename.c_str(), includeWithoutAccents);

    if (trace)
        cout << "FrenchVerbDictionary::init: trie takes "
             << verbTrie.computeMemoryConsumption() << " bytes\n";
}


void
FrenchVerbDictionary::loadConjugationDatabase(
                                const char *conjugationFilename,
                                bool includeWithoutAccents)
                                        throw (logic_error)
{
    if (conjugationFilename == NULL)
        throw invalid_argument("conjugationFilename");

    AutoDoc conjDoc(xmlParseFile(conjugationFilename));
    if (!conjDoc)
        throw logic_error("could not parse " + string(conjugationFilename));

    readConjugation(conjDoc.get(), includeWithoutAccents);
}


void
FrenchVerbDictionary::loadVerbDatabase(
                                const char *verbsFilename,
                                bool includeWithoutAccents)
                                        throw (logic_error)
{
    if (verbsFilename == NULL)
        throw invalid_argument("verbsFilename");

    AutoDoc verbsDoc(xmlParseFile(verbsFilename));
    if (!verbsDoc)
        throw logic_error("could not parse " + string(verbsFilename));

    readVerbs(verbsDoc.get(), includeWithoutAccents);
}


void
FrenchVerbDictionary::readConjugation(xmlDocPtr doc, bool includeWithoutAccents) throw(logic_error)
{
    const bool isItalian = (langCode == "it");

    xmlNodePtr rootNodePtr = xmlDocGetRootElement(doc);

    if (rootNodePtr == NULL)
        throw logic_error("empty conjugation document");

    if (different(rootNodePtr->name, ("conjugation-" + langCode).c_str()))
    {
        string msg = "wrong top node in conjugation document: got "
                     + string((const char *) rootNodePtr->name)
                     + ", expected conjugation-" + langCode;
        throw logic_error(msg);
    }

    for (xmlNodePtr templ = rootNodePtr->xmlChildrenNode;
                        templ != NULL;
                        templ = templ->next)
    {
        if (different(templ->name, "template"))  // ignore junk between tags
            continue;

        string tname = getLatin1XmlProp(templ, "name");
        if (tname.empty())
            throw logic_error("missing template name attribute");

        // The template name is the root and the termination,
        // with a colon in between.  For example, "pla:cer".

        if (tname.find(':') == string::npos)
            throw logic_error("missing colon in template name");

        // The use of the [] operator creates an empty conjugation
        // template spec, to which we keep a reference:

        TemplateSpec &theTemplateSpec = conjugSys[tname];

        // Same idea:

        TemplateInflectionTable &ti = inflectionTable[tname];

        // For each mode (e.g., infinitive, indicative, conditional, etc):
        for (xmlNodePtr mode = templ->xmlChildrenNode;
                            mode != NULL;
                            mode = mode->next)
        {
            if (equal(mode->name, "text") || equal(mode->name, "comment"))  // any text in this node is ignored
                continue;

            Mode theMode = ::convertModeName(mode->name);
            ModeSpec &theModeSpec = theTemplateSpec[theMode];

            // For each tense in the mode:
            for (xmlNodePtr tense = mode->xmlChildrenNode;
                            tense != NULL;
                                tense = tense->next)
            {
                if (equal(tense->name, "text") || equal(tense->name, "comment"))
                    continue;

                Tense theTense = ::convertTenseName(tense->name);
                TenseSpec &theTenseSpec = theModeSpec[theTense];

                // For each person in the tense:
                int personCounter = 0;
                for (xmlNodePtr person = tense->xmlChildrenNode;
                                person != NULL;
                                person = person->next)
                {
                    if (different(person->name, "p"))
                        continue;

                    personCounter++;

                    theTenseSpec.push_back(PersonSpec());
                    PersonSpec &thePersonSpec = theTenseSpec.back();

                    // For each variant for this person:
                    // (Note that most persons of most verbs have only
                    // one variant.)
                    for (xmlNodePtr inf = person->xmlChildrenNode;
                                    inf != NULL;
                                    inf = inf->next)
                    {
                        string variant = getLatin1XmlNodeText(
                                                    doc, inf->xmlChildrenNode);
                        thePersonSpec.push_back(InflectionSpec(variant, true));

                        ModeTensePersonNumber mtpn(
                                reinterpret_cast<const char *>(mode->name),
                                reinterpret_cast<const char *>(tense->name),
                                personCounter,
                                true,
                                isItalian);
                        ti[variant].push_back(mtpn);

                        if (includeWithoutAccents)
                        {
                            // Also include versions where some of all accents are missing.
                            vector<string> unaccentedVariants;
                            formUnaccentedVariants(variant, 0, unaccentedVariants);
                            for (vector<string>::const_iterator it = unaccentedVariants.begin();
                                                                it != unaccentedVariants.end(); ++it)
                            {
                                thePersonSpec.push_back(InflectionSpec(*it, false));
                                mtpn.correct = false;  // 'false' marks this spelling as incorrect.
                                ti[*it].push_back(mtpn);
                            }
                        }
                    }
                }
            }
        }
    }
}


string
FrenchVerbDictionary::getLatin1XmlNodeText(xmlDocPtr doc, xmlNodePtr node)
                                                                throw(int)
{
    xmlChar *s = getString(doc, node);
    if (s == NULL)
        return string();
    return utf8ToLatin1((char *) s);
}


string
FrenchVerbDictionary::getLatin1XmlProp(xmlNodePtr node, const char *propName)
                                                                throw(int)
{
    xmlChar *s = getProp(node, propName);
    if (s == NULL)
        return string();
    return utf8ToLatin1((char *) s);
}


void
FrenchVerbDictionary::readVerbs(xmlDocPtr doc,
                                bool includeWithoutAccents)
                                                throw(logic_error)
{
    xmlNodePtr rootNodePtr = xmlDocGetRootElement(doc);

    if (rootNodePtr == NULL)
        throw logic_error("empty verbs document");

    if (different(rootNodePtr->name, ("verbs-" + langCode).c_str()))
        throw logic_error("wrong top node in verbs document");

    for (xmlNodePtr v = rootNodePtr->xmlChildrenNode; v != NULL; v = v->next)
    {
        if (equal(v->name, "text") || equal(v->name, "comment"))
            continue;

        xmlNodePtr i = v->xmlChildrenNode;
        if (i == NULL || i->xmlChildrenNode == NULL)
            throw logic_error("missing <i> node");

        string infinitive =
                        getLatin1XmlNodeText(doc, i->xmlChildrenNode);
        if (infinitive.empty())
            throw logic_error("empty <i> node");
        size_t lenInfinitive = infinitive.length();
        if (trace) cout << "infinitive='" << infinitive << "'\n";

        if (i->next == NULL)
            throw logic_error("unexpected end after <i> node");

        xmlNodePtr t = i->next->next;
        if (t == NULL)
            throw logic_error("missing <t> node");

        #if 0
        cout << "t=" << t << ", t->xmlChildrenNode=" << t->xmlChildrenNode << "\n";
        if (t->xmlChildrenNode == NULL)
            cout << "  t->next=" << t->next << ", " << (t->next ? getLatin1XmlNodeText(doc, t->next->xmlChildrenNode) : 0) << endl;
        #endif
        string tname = getLatin1XmlNodeText(doc, t->xmlChildrenNode);
        if (tname.empty())
            throw logic_error("empty <t> node");
        if (trace) cout << "  tname='" << tname << "'\n";
        string::size_type posColon = tname.find(':');
        if (posColon == string::npos)
            throw logic_error("missing colon in <t> node");
        if (conjugSys.find(tname) == conjugSys.end())
            throw logic_error("unknown template name: " + tname);

        knownVerbs[infinitive].insert(tname);

        if (includeWithoutAccents)
        {
            // Also include versions where some of all accents are missing.
            vector<string> unaccentedVariants;
            formUnaccentedVariants(infinitive, 0, unaccentedVariants);
            for (vector<string>::const_iterator it = unaccentedVariants.begin();
                                                it != unaccentedVariants.end(); ++it)
            {
                if (trace) cout << "  unaccvar: '" << *it << "'\n";
                knownVerbs[*it].insert(tname);
            }
        }

        // <aspirate-h>: If this verb starts with an aspirate h, remember it:
        if (t->next != NULL && t->next->next != NULL)
            aspirateHVerbs.insert(infinitive);

        // Insert the verb in the trie.
        // A list of template names is associated to each verb in this trie.

        size_t lenTermination = tname.length() - posColon - 1;
        assert(lenTermination > 0);
        assert(lenInfinitive >= lenTermination);

        string verbRadical(infinitive, 0, lenInfinitive - lenTermination);

        insertVerbRadicalInTrie(verbRadical, tname, verbRadical);

        if (includeWithoutAccents)
        {
            // Also include versions where some of all accents are missing.
            vector<string> unaccentedVariants;
            formUnaccentedVariants(verbRadical, 0, unaccentedVariants);
            for (vector<string>::const_iterator it = unaccentedVariants.begin();
                                                it != unaccentedVariants.end(); ++it)
            {
                insertVerbRadicalInTrie(*it, tname, verbRadical);  // pass correct verb radical as 3rd argument
            }
        }
    }

    if (trace)
        cout << "Number of known verbs (" << langCode << "): " << knownVerbs.size() << endl;
}


void
FrenchVerbDictionary::insertVerbRadicalInTrie(
                                    const std::string &verbRadical,
                                    const std::string &tname,
                                    const std::string &correctVerbRadical)
{
    vector<TrieValue> **templateListPtr =
                            verbTrie.getUserDataPointer(verbRadical);
    assert(templateListPtr != NULL);

    // If a new entry was created for 'verbRadical', then the associated
    // user data pointer is null.  Make this pointer point to a new,
    // empty vector of template names.
    //
    if (*templateListPtr == NULL)
        *templateListPtr = new vector<TrieValue>();

    // Associate the given template name to the given verb radical.
    //
    (*templateListPtr)->push_back(TrieValue(tname, correctVerbRadical));
}


FrenchVerbDictionary::~FrenchVerbDictionary()
{
    iconv_close(toLatin1);
    iconv_close(toUTF8);
}


const TemplateSpec *
FrenchVerbDictionary::getTemplate(const string &templateName) const
{
    ConjugationSystem::const_iterator it = conjugSys.find(templateName);
    if (it == conjugSys.end())
        return NULL;
    return &it->second;
}


ConjugationSystem::const_iterator
FrenchVerbDictionary::beginConjugSys() const
{
    return conjugSys.begin();
}


ConjugationSystem::const_iterator
FrenchVerbDictionary::endConjugSys() const
{
    return conjugSys.end();
}


const std::set<std::string> &
FrenchVerbDictionary::getVerbTemplateSet(const char *infinitive) const
{
    static const std::set<std::string> emptySet;
    if (infinitive == NULL)
        return emptySet;
    VerbTable::const_iterator it = knownVerbs.find(infinitive);
    if (it == knownVerbs.end())
        return emptySet;
    return it->second;
}


const std::set<std::string> &
FrenchVerbDictionary::getVerbTemplateSet(const string &infinitive) const
{
    return getVerbTemplateSet(infinitive.c_str());
}


VerbTable::const_iterator
FrenchVerbDictionary::beginKnownVerbs() const
{
    return knownVerbs.begin();
}


VerbTable::const_iterator
FrenchVerbDictionary::endKnownVerbs() const
{
    return knownVerbs.end();
}


const std::vector<ModeTensePersonNumber> *
FrenchVerbDictionary::getMTPNForInflection(
                                const std::string &templateName,
                                const std::string &inflection) const
{
    InflectionTable::const_iterator i = inflectionTable.find(templateName);
    if (i == inflectionTable.end())
        return NULL;
    const TemplateInflectionTable &ti = i->second;
    TemplateInflectionTable::const_iterator j = ti.find(inflection);
    if (j == ti.end())
        return NULL;
    return &j->second;
}


/*static*/
Mode
FrenchVerbDictionary::convertModeName(const char *modeName)
{
    Mode mode = INVALID_MODE;
    if (modeName == NULL)
        ;
    else if (strcmp(modeName, "infinitive") == 0)
        mode = INFINITIVE_MODE;
    else if (strcmp(modeName, "indicative") == 0)
        mode = INDICATIVE_MODE;
    else if (strcmp(modeName, "conditional") == 0)
        mode = CONDITIONAL_MODE;
    else if (strcmp(modeName, "subjunctive") == 0)
        mode = SUBJUNCTIVE_MODE;
    else if (strcmp(modeName, "imperative") == 0)
        mode = IMPERATIVE_MODE;
    else if (strcmp(modeName, "participle") == 0)
        mode = PARTICIPLE_MODE;
    else if (strcmp(modeName, "gerund") == 0)
        mode = GERUND_MODE;

    if (mode == INVALID_MODE)
    {
        if (trace) cout << "modeName='" << modeName << "'" << endl;
        assert(!"Invalid mode");
    }

    return mode;
}


/*static*/
Tense
FrenchVerbDictionary::convertTenseName(const char *tenseName)
{
    Tense tense = INVALID_TENSE;
    if (tenseName == NULL)
        ;
    else if (strcmp(tenseName, "infinitive-present") == 0)
        tense = PRESENT_TENSE;
    else if (strcmp(tenseName, "present") == 0)
        tense = PRESENT_TENSE;
    else if (strcmp(tenseName, "imperfect") == 0)
        tense = IMPERFECT_TENSE;
    else if (strcmp(tenseName, "future") == 0)
        tense = FUTURE_TENSE;
    else if (strcmp(tenseName, "simple-past") == 0)
        tense = PAST_TENSE;
    else if (strcmp(tenseName, "imperative-present") == 0)
        tense = PRESENT_TENSE;
    else if (strcmp(tenseName, "present-participle") == 0)
        tense = PRESENT_TENSE;
    else if (strcmp(tenseName, "past-participle") == 0)
        tense = PAST_TENSE;
    else if (strcmp(tenseName, "past") == 0)
        tense = PAST_TENSE;
    else if (strcmp(tenseName, "present-gerund") == 0)
        tense = PRESENT_TENSE;

    if (tense == INVALID_TENSE)
    {
        if (trace) cout << "tenseName='" << tenseName << "'" << endl;
        assert(!"Invalid tense");
    }

    return tense;
}


void
FrenchVerbDictionary::deconjugate(const string &conjugatedVerb,
                                std::vector<InflectionDesc> &results)
{
    verbTrie.setDestination(&results);
    (void) verbTrie.get(conjugatedVerb);
    verbTrie.setDestination(NULL);
}


/*virtual*/
void
FrenchVerbDictionary::VerbTrie::onFoundPrefixWithUserData(
                        const string &conjugatedVerb,
                        string::size_type index,
                        const vector<TrieValue> *templateList) const throw()
{
    if (trace)
        cout << "VerbTrie::onFoundPrefixWithUserData: start: conjugatedVerb='" << conjugatedVerb << "'" << endl;

    assert(templateList != NULL);
    if (results == NULL)
        return;

    const string radical(conjugatedVerb, 0, index);
    const string term(conjugatedVerb, index);

    /*
        'templateList' contains the names of conjugated templates that might
        apply to the conjugated verb.  We check each of them to see if there
        is one that accepts the given termination 'term'.
    */
    for (vector<TrieValue>::const_iterator i = templateList->begin();
                                           i != templateList->end(); i++)
    {
        const TrieValue &trieValue = *i;
        const string &tname = trieValue.templateName;
        const TemplateInflectionTable &ti =
                                fvd.inflectionTable.find(tname)->second;
        TemplateInflectionTable::const_iterator j = ti.find(term);
        if (j == ti.end())
            continue;  // template 'tname' does not accept termination 'term'

        // template 'tname' accepts 'term', so we produce some results.

        string templateTerm(tname, tname.find(':') + 1);
            // termination of the infinitive form

        const vector<ModeTensePersonNumber> &v = j->second;
            // list of mode-tense-person combinations that can correspond
            // to the conjugated verb's termination

        for (vector<ModeTensePersonNumber>::const_iterator k = v.begin();
                                                    k != v.end(); k++)
        {
            const ModeTensePersonNumber &mtpn = *k;

            string infinitive = trieValue.correctVerbRadical + templateTerm;
                // The infinitive of the conjugated verb is formed from its
                // (correct) radical part and from the termination of the template name.
                // Correct means with the proper accents. This allows the user
                // to type "etaler" without the acute accent on the first "e"
                // and obtain the conjugation for the correct verb, which has
                // that accent.

            if (trace)
                cout << "VerbTrie::onFoundPrefixWithUserData: radical='"
                    << radical << "', templateTerm='" << templateTerm
                    << "', tname='" << tname
                    << "', correctVerbRadical='" << trieValue.correctVerbRadical
                    << "', mtpn=("
                    << mtpn.mode << ", "
                    << mtpn.tense << ", "
                    << (unsigned) mtpn.person << ", "
                    << mtpn.plural << ", "
                    << mtpn.correct << ")\n";

            results->push_back(InflectionDesc(infinitive, tname, mtpn));
                // the InflectionDesc object is an analysis of the
                // conjugated verb
        }
    }
}


/*static*/
const char *
FrenchVerbDictionary::getModeName(Mode m)
{
    if (int(m) < int(INFINITIVE_MODE) || int(m) > int(GERUND_MODE))
        return NULL;

    static const char *names[] =
    {
        "infinitive", "indicative", "conditional",
        "subjunctive", "imperative", "participle",
        "gerund"
    };

    return names[int(m) - 1];
}


/*static*/
const char *
FrenchVerbDictionary::getTenseName(Tense t)
{
    if (int(t) < int(PRESENT_TENSE) || int(t) > int(FUTURE_TENSE))
        return NULL;

    static const char *names[] =
    {
        "present", "past", "imperfect", "future"
    };

    return names[int(t) - 1];
}


string
FrenchVerbDictionary::tolowerLatin1(const string &latin1String) const
{
    string result;
    for (string::size_type len = latin1String.length(), i = 0; i < len; i++)
        result += latin1TolowerTable[(unsigned char) latin1String[i]];
    return result;
}


string
FrenchVerbDictionary::latin1ToUTF8(const string &latin1String) const throw(int)
{
    size_t len = latin1String.length();
    size_t inbytesleft = len + 1;
    size_t outbytesleft = len * 2 + 1;
    char *inbuf = strcpy(new char[inbytesleft], latin1String.c_str());
    char *outbuf = new char[outbytesleft];

    ICONV_CONST char *in = inbuf;
        /*  ICONV_CONST is defined by iconv.m4 to 'const' if the 2nd
            argument of iconv() requires const, as is the case under
            FreeBSD 6.1 with GCC 3.4.4 [2006-07-10].
        */
    char *out = outbuf;
    if (iconv(toUTF8, &in, &inbytesleft, &out, &outbytesleft) == (size_t) -1)
    {
        int e = errno;
        delete [] inbuf;
        delete [] outbuf;
        throw e;
    }

    string result = outbuf;
    delete [] inbuf;
    delete [] outbuf;
    return result;
}


string
FrenchVerbDictionary::utf8ToLatin1(const string &utf8String) const throw(int)
{
    size_t len = utf8String.length();
    size_t inbytesleft = len + 1;
    size_t outbytesleft = len + 1;
    char *inbuf = strcpy(new char[inbytesleft], utf8String.c_str());
    char *outbuf = new char[outbytesleft];

    ICONV_CONST char *in = inbuf;
    char *out = outbuf;
    if (iconv(toLatin1, &in, &inbytesleft, &out, &outbytesleft) == (size_t) -1)
    {
        int e = errno;
        delete [] inbuf;
        delete [] outbuf;
        throw e;
    }

    string result = outbuf;
    delete [] inbuf;
    delete [] outbuf;
    return result;
}


void
FrenchVerbDictionary::utf8ToLatin1(vector<InflectionDesc> &vec) const throw(int)
{
    for (vector<InflectionDesc>::iterator it = vec.begin();
                                        it != vec.end(); it++)
        (*it).infinitive = utf8ToLatin1((*it).infinitive);

    // The 'templateName' field is in ASCII and does not need to be converted.
}


void
FrenchVerbDictionary::utf8ToLatin1(
                        vector<vector<string> > &vec) const throw(int)
{
    for (vector<vector<string> >::iterator i = vec.begin();
                                        i != vec.end(); i++)
        for (vector<string>::iterator j = (*i).begin();
                                        j != (*i).end(); j++)
            *j = utf8ToLatin1(*j);
}



/*static*/
string
FrenchVerbDictionary::getRadical(
                        const string &infinitive,
                        const string &templateName) throw(logic_error)
{
    string::size_type posColon = templateName.find(':');
    if (posColon == string::npos)
        throw logic_error("no colon found in template name");

    string::size_type lenSuffix = templateName.length() - posColon - 1;
    string::size_type lenInfPrefix = infinitive.length() - lenSuffix;
    return string(infinitive, 0, lenInfPrefix);
}


/*static*/
bool
FrenchVerbDictionary::generateTense(const string &radical,
                                const TemplateSpec &templ,
                                Mode mode,
                                Tense tense,
                                vector< vector<string> > &dest,
                                bool includePronouns,
                                bool aspirateH,
                                bool isItalian) throw()
{
    if (templ.find(mode) == templ.end())
        return false;

    const ModeSpec &modeSpec = templ.find(mode)->second;

    if (modeSpec.find(tense) == modeSpec.end())
        return false;

    const TenseSpec &tenseSpec = modeSpec.find(tense)->second;

    if (mode != INDICATIVE_MODE
            && mode != CONDITIONAL_MODE
            && mode != SUBJUNCTIVE_MODE)
        includePronouns = false;

    for (TenseSpec::const_iterator p = tenseSpec.begin();
                                    p != tenseSpec.end(); p++)
    {
        dest.push_back(vector<string>());
        for (PersonSpec::const_iterator i = p->begin(); i != p->end(); i++)
        {
            // Do not return spellings that are marked incorrect.
            // They are in the knowledge base only to allow
            // error-tolerant searches.
            //
            if (!(*i).isCorrect)
                continue;

            string pronoun;  // no pronoun by default

            string v = radical + (*i).inflection;

            if (includePronouns)
            {
                size_t noPers = p - tenseSpec.begin();
                switch (noPers)
                {
                    case 0:
                    if (isItalian)
                        pronoun = "io ";
                    else
                    {
                        bool elideJe = false;
                        if (!aspirateH)
                        {
                            char init = (v.empty() ? '\0' : v[0]);
                            bool isVowelOrH = (strchr(
                                    "aeiouyhAEIOUYH",
                                                                init) != NULL);
                            if (isVowelOrH)
                                elideJe = true;
                        }
                        pronoun = (elideJe ? "j'" : "je ");
                    }
                    break;
                    case 1: pronoun = "tu "; break;
                    case 2: pronoun = (isItalian ? "egli " : "il "); break;
                    case 3: pronoun = (isItalian ? "noi "  : "nous "); break;
                    case 4: pronoun = (isItalian ? "voi "  : "vous "); break;
                    case 5: pronoun = (isItalian ? "essi " : "ils "); break;
                }

                if (mode == SUBJUNCTIVE_MODE)
                {
                    const char *conj;
                    if (isItalian)
                        conj = "che ";
                    else if (noPers == 2 || noPers == 5)
                        conj = "qu'";
                    else
                        conj = "que ";
                    pronoun = conj + pronoun;
                }
            }

            dest.back().push_back(pronoun + v);
        }
    }

    return true;
}


bool FrenchVerbDictionary::isVerbStartingWithAspirateH(
                                const std::string &infinitive) const throw()
{
    return aspirateHVerbs.find(infinitive) != aspirateHVerbs.end();
}
